/*
   netrik -- The ANTRIK Internet Viewer
   Copyright (C) Olaf D. Buddenhagen AKA antrik, et al (see AUTHORS)
   Published under the GNU GPL; see LICENSE for details.
*/
/*
 * pager.c -- Interactive viewing module.
 *
 * (C) 2001 antrik
 *
 * Displays output page and handles user commands.
 */
#include <curses.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>

#include "cfg.h"
#include "layout.h"
#include "links.h"
#include "page.h"
#include "pager.h"
#include "render.h"
#include "screen.h"

enum Search_type {
   FIND_FIRST,
   FIND_LAST
};

static int active_start(const struct Page *page, int pos);    /* find first active line */
static int active_end(const struct Page *page, int pos);    /* find last active line */

static int find_link(const struct Page *page, int start_line, int end_line, int start_link, int end_link, enum Search_type search_type);    /* find link matching criteria */

static void activate_link(struct Page *page, int change_link);    /* set active link */
static void scroll_to(struct Page *page, int new_line);    /* adjust visible part of page */
static void activate_anchor(struct Page *page);    /* show active anchor */

#define PAGE_INVALID ((unsigned int)-1>>1)    /* biggest possible positive int */
static int	old_line;    /* page position visible up to now */

/* return first line (page coordinates) where links can be activated with the
 * current "pager_pos" plus the relative "pos" */
static int active_start(page, pos)
const struct Page	*page;
int			pos;
{
   const int	screen_start=page->pager_pos+pos;

   if(screen_start > 0)
      return screen_start+cfg.link_margin;
   else
      return 0;
}

/* return pos *after* last line (page coordinates) where links can be
 * activated with the current "pager_pos" plus the relative "pos" */
static int active_end(page, pos)
const struct Page	*page;
int			pos;
{
   const int	page_end=page->layout->item_tree->y_end;
   const int	screen_end=page->pager_pos+pos+LINES;

   if(screen_end < page_end)
      return screen_end-cfg.link_margin;
   else
      return page_end;
}

/* 
 * Find a link that matches certain criteria.
 *
 * In particular, the link's position on the page has to be inside a specified
 * range of lines, and the link number also has to be inside a specified range.
 * As usual, the ranges are specified by pairs of two values specifying the
 * first valid line/link, and the *first invalid* line/link, i.e. one behind
 * the last valid one.
 *
 * Not all criteria have to be specified; any one can be omitted (by passing
 * -1), meaning there is no limitation in that direction.
 *
 * The "search_type" flag decides whether the first link (list order) which
 * lies inside both ranges is returned, or the last one.
 */

static int find_link(page, start_line, end_line, start_link, end_link, search_type)
const struct Page	*page;    /* page to search */
int			start_line, end_line;    /* range of allowed lines */
int			start_link, end_link;    /* range of allowed link numbers */
enum Search_type	search_type;    /* take first match or last match (list direction) */
{
   int			dir;    /* search direction */
   enum Search_type	match_type;    /* take first or last match (in search direction) */
   int			first, last;    /* first and last link to test (in search order!) */
   int			tested;    /* currently tested link */
   int			candidate;    /* matching link */

   if(start_link<0)    /* no "start_link" specified -> start with first in list */
      start_link=0;

   if(end_link<0) {    /* no end link specified -> serach forwards */
      dir=1;
      first=start_link;
      last=page->layout->links->count;    /* no end link, so scan till last */
      match_type=search_type;    /* first found is also first in range */
   } else {    /* end link specified -> search backwards */
      dir=-1;
      first=end_link-1;    /* -1 because "end_link" not to be included, so start at the one before */
      last=start_link-1;    /* -1 because "start_link" is to be included, so stop at the one before */
      match_type= search_type==FIND_FIRST ? FIND_LAST : FIND_FIRST;    /* need to reverse as first found is last in range */
   }

   candidate=-1;    /* nothing found yet */
   for(tested=first; tested!=last; tested+=dir) {    /* all links in range */
      const struct Link	*link=get_link(page->layout, tested);    /* current link's data */

      const int		in_lower_bound= link->y_start >= start_line;
      const int		in_upper_bound= (unsigned)link->y_end <= (unsigned)end_line;    /* unsigned so -1 for "end_line" is treated as big positive number, i.e. no limit */

      if(in_lower_bound && in_upper_bound) {    /* inside line number range */
	 candidate=tested;    /* -> take it */
	 if(match_type==FIND_FIRST)    /* only find first -> we are ready (otherwise continue scanning) */
	    break;
      } else {    /* not in range */
	 if((dir==1 && !in_upper_bound) || (dir==-1 && !in_lower_bound))    /* moving away from valid range -> stop scanning */
	    break;
      }
   }    /* for all links in range */

   return candidate;
}

/*
 * set active link
 *
 * Also highlights the new link and redraws the affected screen area.
 *
 * If some link was alredy active, it is first deactivated before activating
 * the new one.
 *
 * If change_link is -1, deactivate link (no link is active afterwards).
 */

static void activate_link(page, change_link)
struct Page	*page;
int		change_link;
{
   const int			link_num=change_link>=0?change_link:page->active_link;    /* link to activate or deactivate (normally, change (activate) given link; if -1, change presently active one (deactivate)) */

   if(link_num>=0) {    /* actually have work to do (not trying to deactivate when there was no link active anyways) */
      const struct Link_ptr	*list_entry=&page->layout->links->link[link_num];    /* entry in link list pointing to link */
      const struct Item		*item=list_entry->item;    /* item containing link */
      const struct String	*string=item->data.string;    /* string containing link */
      const struct Link		*link=&string->link[list_entry->num];    /* link data */

      int	x_start, y_start, x_end, y_end;    /* screen area that needs to be redrawn (affected by link) */
      int	page_x, page_y;    /* starting position of redrawn area from page */

      /* scroll page so link will be on screen */
      if(change_link>=0) {    /* activating link, not deactivating (deactivating a link not on screen is OK) */
	 const int	first_line=active_start(page, 0);    /* first line in which links may be activated */
	 const int	last_line=active_end(page, 0);    /* (after) last line where links may be activated */

	 int	scroll_lines=0;

	 if(link->y_end > last_line)    /* link ends behind active screen area */
	    scroll_lines=link->y_end-last_line;    /* -> scroll as many lines as necessary so it fits */
	 if(link->y_start < first_line)    /* starts before active area */
	    scroll_lines=link->y_start-first_line;    /* -> scroll backwards (negative scroll_lines) NOTE: this may undo previous condition, if link spans more lines than active area... */

	 if(scroll_lines) {    /* have to scroll */
	    if(abs(scroll_lines)<LINES)    /* need to scroll less then screenful */
	       scroll_to(page, page->pager_pos+scroll_lines);    /* -> scroll just as far as necessary so link fits */
	    else    /* scroll very far */
	       scroll_to(page, link->y_start-cfg.link_margin);    /* -> scroll so that link will appear just at the beginning of active screen area (this is identical to scrolling just as far as necessary when scrolling back, but different when scrolling forward) */
	 }
      }    /* scroll */

      if(change_link>=0)    /* activating link, not deactivating */
	 if(page->active_link>=0)    /* some link is already active */
	    activate_link(page, -1);    /* -> first deactivate the old one */

      highlight_link(page->layout, link_num, change_link>=0?0:1);    /* highlight the link, or unhighlight if deactivating */

      /* coordinates of updated area */
      if(link->y_end-link->y_start==1) {    /* link is on one line -> only redraw the affected line part */
	 page_x=link->x_start;
	 x_start=link->x_start;    /* no x-scrolling yet... */
	 x_end=link->x_end;
      } else {    /* link spans multiple lines -> completely redraw all lines that contain some part of the link */
	 /* full width of item (whole lines) */
	 page_x=item->x_start;
	 x_start=item->x_start;
	 x_end=item->x_end;
      }
      page_y=link->y_start;
      y_start=link->y_start-page->pager_pos;    /* screen coordinates of link depend on pager position */
      y_end=link->y_end-page->pager_pos;

      if(x_start<COLS && x_end>0 && y_start<LINES && y_end>0) {    /* some part visible */
	 /* truncate to visible area */
	 if(x_start<0) {
	    page_x-=x_start;
	    x_start=0;
	 }
	 if(y_start<0) {
	    page_y-=y_start;
	    y_start=0;
	 }
	 if(x_end>COLS)
	    x_end=COLS;
	 if(y_end>LINES)
	    y_end=LINES;

	 render(page->layout, page_x, page_y, x_start, y_start, x_end-x_start, y_end-y_start, 0);    /* update area affected by link */
      }

      page->active_link=change_link;
   }    /* anything to do */
}

/* 
 * move visible page area so "new_line" will be first line visible on screen
 *
 * If this isn't possible because of page boundaries, scroll as far as
 * possible.
 */

static void scroll_to(page, new_line)
struct Page	*page;
int		new_line;    /* page position to move to */
{
   int		scroll_lines;    /* number of lines to scroll */

   /* page boundaries */
   if(new_line>page->layout->item_tree->y_end-LINES)    /* page ends before end of screen -> adjust */
      new_line=page->layout->item_tree->y_end-LINES;
   if(new_line<0)    /* page begins after start of screen -> adjust (may undo previous condition, if page length < screen length!) */
      new_line=0;

   scroll_lines=new_line-old_line;

   if(abs(scroll_lines)<=LINES) {    /* move not more than a screenfull -> scroll */
      move(0,0);
      if(insdelln(-scroll_lines)==ERR) {
	 endwin();
	 fprintf(stderr, "curses error: can't scroll (in function scroll_to)\n");
	 exit(2);
      }
      
      if(scroll_lines<0)    /* scroll back */
	 render(page->layout, 0, new_line, 0, 0, COLS, -scroll_lines, 0);    /* render new first line(s) */
      else    /* scroll forward */
	 render(page->layout, 0, new_line+LINES-scroll_lines, 0, LINES-scroll_lines, COLS, scroll_lines, 0);    /* render new last line(s) */

   } else {    /* move more than screenful (or PAGE_INVALID) -> repaint */
      erase();
      render(page->layout, 0, new_line, 0, 0, COLS, LINES, 0);    /* fill whole screen */
   }
   
   old_line=new_line;    /* save new position */
   page->pager_pos=new_line;

   if(page->active_link>=0) {    /* some link is active -> may have to deactivate */
      const struct Link	*link=get_link(page->layout, page->active_link);    /* current link's data */

      if(link->y_start < active_start(page, 0) || link->y_end > active_end(page, 0))    /* moved out of legal area for active links */
	 activate_link(page, -1);    /* -> deactivate */
   }
}

/*
 * Show active anchor.
 *
 * Scroll so that the anchor will be in the desired screen line (determined by
 * "cfg.anchor_offset"); draw marks at all screen lines spanned by the anchor.
 * When called while no "active_anchor" set, clear anchor marks.
 *
 * First the anchor position is retrieved and the page scrolled to the optimal
 * position. Afterwards, the real anchor position on the screen is calculated,
 * and the marks set directly to the screen. (Not added to "item_tree".)
 * Clearing is done by rerendering the affected area.
 */

static void activate_anchor(page)
struct Page	*page;
{
   static int	anchor;    /* affected anchor */
   int		start_line;    /* screen line containing anchor start */
   int		end_line;    /* screen lines after anchor end */

   if(page->active_anchor>=0)    /* there is an anchor to activate -> take it (otherwise use the one stored before) */
      anchor=page->active_anchor;

   /* get screen position of anchor */
   {
      const struct Item	*anchor_item=page->layout->anchors->anchor_item[anchor];

      if(page->active_anchor>=0)    /* activating anchor -> first need to scroll */
	 scroll_to(page, anchor_item->y_start-(cfg.anchor_offset?LINES/cfg.anchor_offset:cfg.link_margin));    /* scroll so anchor will appear "1/anchor_offset" below screen top, or "link_margin" below screen top if no "anchor_offset" */

      start_line=anchor_item->y_start - page->pager_pos;
      end_line=anchor_item->y_end - page->pager_pos;

      if(end_line==start_line)    /* empty anchor -> mark one line */
	 ++end_line;

      if(start_line<0)    /* starts before screen top -> truncate */
	 start_line=0;

      if(end_line>LINES)
	 end_line=LINES;
   }

   if(page->active_anchor>=0) {    /* activate -> draw marks */
      int	line;

#ifdef DEBUG
      if(start_line>=LINES || end_line<=0) {
	 endwin();
	 fprintf(stderr, "internal error: trying to activate anchor not on screen\n");
	 exit(100);
      }
#endif

      for(line=start_line; line<end_line; ++line) {
	 set_color(color_map[TM_SYS]);
	 mvprintw(line, COLS-1, "%c", '<');
      }
   } else    /* deactivate -> rerender area containing marks */
      if(end_line>start_line)    /* anything to render (not scrolled out of screen) */
	 render(page->layout, COLS-1, start_line+page->pager_pos, COLS-1, start_line, 1, end_line-start_line, 1);    /* overpaint */
}

/*
 * Interactive pager.
 *
 * Until some command requirering returning from pager (following link,
 * pressing 'q' etc.) issued: Display part of output page on curses screen;
 * read input key; move visible part according to pressed key function.
 *
 * When some command was issued that requires quitting the pager and returning
 * to the main program, return a flag indicating what was the cause, so main()
 * knows what to do.
 */

enum Pager_ret display(page)
struct Page	*page;
{
   enum Pager_ret	quit=RET_NO;

   sigaction(SIGWINCH, &winch_action, NULL);    /* restore default handler to force curses to reinitialize it's own handler (readline() breaks it) */

   /* create initial pager screen */

   if(cbreak()==ERR) {    /* readline resets this... */
      endwin();
      fprintf(stderr, "curses error while (re)starting pager");
      exit(2);
   }

   old_line=PAGE_INVALID;    /* force repaint */

   if(page->active_anchor>=0)    /* anchor requested -> jump to it */
      activate_anchor(page);
   else    /* no anchor -> just go to current page position */
      scroll_to(page, page->pager_pos);    /* display visible part of page */

   if(page->active_link>=0)    /* some link active -> force highlighting (necessary after reloading from history) */
      activate_link(page, page->active_link);

   /* keyboard loop */
   do {    /* until quit */
      int	cursor_x, cursor_y;

      if(page->active_link >= 0) {
	 const struct Link	*link=get_link(page->layout, page->active_link);    /* current active link's data */

	 cursor_x=link->x_start;
	 cursor_y=link->y_start-page->pager_pos;
      } else
	 cursor_x=cursor_y=0;

      /* hack: ignore wrong cursor coordinates (until SIGWINCH will be handled properly) */
      if(cursor_x>=COLS)
	 cursor_x=COLS-1;
      if(cursor_y>=LINES)
	 cursor_y=LINES-1;

#ifdef DEBUG
      if((unsigned)cursor_x>=(unsigned)COLS || (unsigned)cursor_y>=(unsigned)LINES) {
	 endwin();
	 fprintf(stderr, "internal error: trying to set cursor outside screen (x: %d; y: %d)\n", cursor_x, cursor_y);
	 exit(100);
      }
#endif

      switch(mvgetch(cursor_y, cursor_x)) {

	 /* one line down */
	 case 'j':
	    scroll_to(page, page->pager_pos+1);
	    break;

	 /* one line up */
	 case 'k':
	    scroll_to(page, page->pager_pos-1);
	    break;

	 /* two lines down */
	 case KEY_DC:    /* DEL */
	    scroll_to(page, page->pager_pos+2);
	    break;

	 /* two lines up */
	 case KEY_IC:    /* INS */
	    scroll_to(page, page->pager_pos-2);
	    break;

	 /* screen forward */
	 case 'F'-'@':    /* ^f */
	 case ' ':
	    scroll_to(page, page->pager_pos+LINES);
	    break;

	 /* screen backward */
	 case 'B'-'@':    /* ^b */
	    scroll_to(page, page->pager_pos-LINES);
	    break;

	 /* half screen forward */
	 case 'D'-'@':    /* ^d */
	 case KEY_NPAGE:
	    scroll_to(page, page->pager_pos+LINES/2);
	    break;

	 /* half screen backward */
	 case 'U'-'@':    /* ^u */
	 case KEY_PPAGE:
	    scroll_to(page, page->pager_pos-LINES/2);
	    break;

	 /* page top */
	 case 'g':
	 case KEY_HOME:
	    scroll_to(page, 0);
	    break;

	 /* page bottom */
	 case 'G':
	 case KEY_END:
	    scroll_to(page, page->layout->item_tree->y_end-LINES);
	    break;

	 /* down/next link */
	 case 'J':
	 case KEY_DOWN: {
	    int	new_link;    /* link that will be activated */

	    new_link=find_link(page, active_start(page, 0), active_end(page, 1), page->active_link+1, -1, FIND_FIRST);    /* find first link in the lines active now or after scrolling one line, after the current active link if any */

	    if(new_link>=0)    /* some new link found */
	       activate_link(page, new_link);    /* -> activate it */
	    else    /* nothing found */
	       scroll_to(page, page->pager_pos+1);    /* -> just scroll a line forward */

	    break;
	 }

	 /* up/next link */
	 case 'K':
	 case KEY_UP: {
	    int	new_link;    /* link that will be activated */

	    if(page->active_link >= 0)    /* some link is active */
	       new_link=find_link(page, active_start(page, -1), -1, -1, page->active_link, FIND_LAST);    /* -> find last link before the current one, in the lines active now or after scrolling one line */
	    else    /* no link active yet */
	       new_link=find_link(page, active_start(page, -1), active_start(page, 0), -1, -1, FIND_LAST);    /* -> find last link in the line(s) that will become active after scrolling one line */

	    if(new_link>=0)    /* some new link found */
	       activate_link(page, new_link);    /* -> activate it */
	    else    /* nothing found */
	       scroll_to(page, page->pager_pos-1);    /* -> just scroll a line back */

	    break;
	 }

	 /* activate first link on screen */
	 case 'H':
	    if(page->active_link >= 0)    /* have some active link */
	       activate_link(page, find_link(page, active_start(page, 0), active_end(page, 0), -1, page->active_link+1, FIND_FIRST));    /* -> activate first link in current active screen area; search backward from current active link (include current one, as it might be already the first) */
	    else    /* no link active */
	       activate_link(page, find_link(page, active_start(page, 0), active_end(page, 0), -1, -1, FIND_FIRST));    /* -> just activate first link in current active screen area */
	    break;

	 /* activate last link on screen */
	 case 'L':
	    activate_link(page, find_link(page, active_start(page, 0), active_end(page, 0), page->active_link, -1, FIND_LAST));    /* activate last link in current active screen area; start serach from current active link, if any (include current one, as it might be already the last) */
	    break;

	 /* activate first link in second screen half */
	 case 'M': {
	    int	new_link;    /* link that will be activated */

	    new_link=find_link(page, page->pager_pos+LINES/2, active_end(page, 0), -1, page->active_link, FIND_FIRST);    /* find first link in second screen half; search backwards from current active link, if any */
	    if(new_link==-1 && page->active_link >= 0)    /* nothing found, but some link was active already => maybe nothing found because we have searched in wrong direction */
	       new_link=find_link(page, page->pager_pos+LINES/2, active_end(page, 0), page->active_link, -1, FIND_FIRST);    /* -> try searching forward from current active link */

	    if(new_link>=0)    /* something found */
	       activate_link(page, new_link);    /* -> activate */
	    break;
	 }

	 /* activate first link in next line having links (or scroll one line) */
	 case '+':
	 case '=': {
	    int	new_link;

	    if(page->active_link >= 0)    /* some link active */
	       new_link=find_link(page, get_link(page->layout, page->active_link)->y_start+1, active_end(page, 1), page->active_link+1, -1, FIND_FIRST);    /* -> find first link after current one that starts in next line or behind (up to the last line becoming active when scrolling one line) */
	    else
	       new_link=find_link(page, active_start(page, 0), active_end(page, 1), -1, -1, FIND_FIRST);    /* find first link in area active now or after scrolling one line */

	    if(new_link>=0)    /* something found */
	       activate_link(page, new_link);    /* -> activate */
	    else    /* no new link found */
	       scroll_to(page, page->pager_pos+1);    /* -> just scroll one line */

	    break;
	 }

	 /* activate first link in previous line having links (or scroll one line) */
	 case '-': {
	    int	new_link;

	    /* first find some link in previous line having any links */
	    if(page->active_link >= 0)    /* some link active */
	       new_link=find_link(page, active_start(page, -1), get_link(page->layout, page->active_link)->y_start, -1, page->active_link, FIND_LAST);    /* -> find last link before current one that starts in previous line or before (up to the first line becoming active when scrolling one line) */
	    else
	       new_link=find_link(page, active_start(page, -1), active_start(page, 0), -1, -1, FIND_LAST);    /* find last link in lines newly becoming active after scrolling one line */

	    if(new_link>=0) {    /* something found */
	       new_link=find_link(page, get_link(page->layout, new_link)->y_start, -1, -1, new_link+1, FIND_FIRST);    /* -> now get first link in that line */
	       activate_link(page, new_link);
	    } else    /* no new link found */
	       scroll_to(page, page->pager_pos-1);    /* -> just scroll one line */

	    break;
	 }

	 /* activate first link starting on line */
	 case '^':
	 case 'A'-'@':    /* ^A */
	    if(page->active_link >= 0)    /* some link active */
	       activate_link(page, find_link(page, get_link(page->layout, page->active_link)->y_start, -1, -1, page->active_link+1, FIND_FIRST));    /* -> activate first link that starts in the line in which the current active link starts, going backwards from current link (including current one) */
	    break;

	 /* activate first link in line */
	 case '0':
	    if(page->active_link >= 0)    /* some link active */
	       activate_link(page, find_link(page, -1, get_link(page->layout, page->active_link)->y_start, -1, page->active_link, FIND_LAST)+1);    /* -> find last link ending *before* current line, and then activate next one (which is the first link starting and/or ending on current line) */
	    break;

	 case '$':
	    if(page->active_link >= 0) {    /* some link active */
	       int	new_link;

	       new_link=find_link(page, get_link(page->layout, page->active_link)->y_start+1, -1, page->active_link, -1, FIND_FIRST)-1;    /* find first link starting *after* current line, and get previous one (which is the last one starting on current line) */
	       if(new_link>=0)    /* something found */
		  activate_link(page, new_link);    /* -> activate */
	       else    /* no links starting after current line */
		  activate_link(page, page->layout->links->count-1);    /* -> activate last link on page (which is also last on current line) */
	    }
	    break;

	 /* activate first link on page */
	 case KEY_BACKSPACE:
	    if(page->layout->links->count)    /* any links on page */
	       activate_link(page, 0);    /* -> activate first */
	    break;

	 /* activate next link */
	 case 'I'-'@': {    /* Tab (^I) */
	    int	new_link;

	    new_link=find_link(page, active_start(page, 0), -1, page->active_link+1, -1, FIND_FIRST);    /* find first link on screen or behind, after current active one if any */
	    if(new_link>=0)
	       activate_link(page, new_link);
	    break;
	 }

	 /* activate previous link */
	 case 'p': {
	    int	new_link;

	    if(page->active_link >= 0)    /* some link active */
	       new_link=find_link(page, -1, -1, -1, page->active_link, FIND_LAST);    /* -> get previous link */
	    else    /* no link active */
	       new_link=find_link(page, -1, active_start(page, 0), -1, -1, FIND_LAST);    /* find last link before screen start */

	    if(new_link>=0)
	       activate_link(page, new_link);
	    break;
	 }

	 /* follow active link */
	 case '\r':    /* <return> */
	    if(page->active_link>=0)    /* some link active -> follow */
	       quit=RET_LINK;
	    else
	       flash();
	    break;

	 /* show link URL */
	 case 'u':
	    if(page->active_link>=0)    /* some link active */
	       quit=RET_LINK_URL;
	    else
	       flash();
	    break;

	 /* show absolute link target URL */
	 case 'U':
	    if(page->active_link>=0)    /* some link active */
	       quit=RET_ABSOLUTE_URL;
	    else
	       flash();
	    break;

	 /* show current page URL */
	 case 'c':
	    quit=RET_URL;
	    break;

	 /* reload page */
	 case 'R'-'@':    /* ^r */
	    if(page->url->proto.type!=PT_INTERNAL)    /* normal page */
	       quit=RET_HISTORY;    /* reload (same) page */
	    else    /* stdin -> can't reload */
	       flash();
	    break;

	 /* back to previous URL */
	 case 'b':
	 case KEY_LEFT:
	    if(page_list.pos > 0) {    /* any URL(s) to go back */
	       --page_list.pos;
	       quit=RET_HISTORY;
	    } else
	       flash();
	    break;

	 /* forward to next URL (after 'b' or other commands going back in history) */
	 case 'f':
	 case KEY_RIGHT:
	    if(page_list.pos < page_list.count-1) {    /* can go forward (not at last entry) */
	       ++page_list.pos;
	       quit=RET_HISTORY;
	    } else
	       flash();
	    break;

	 /* back to URL before last absolute link/load */
	 case 'B':
	    if(page_list.pos > 0) {    /* any URL(s) to go back */
	       int	pos;

	       for(pos=page_list.pos-1; pos>0; --pos)    /* traverse history backwards (stop at first entry) */
		  if(page_list.page[pos+1]->url->absolute)    /* found matching entry (a page before one with absolute URL) */
		     break;
	       page_list.pos=pos;    /* take found entry (or first entry, if nothing found) */

	       quit=RET_HISTORY;
	    } else    /* already at first entry */
	       flash();
	    break;

	 /* forward to URL before next absolute link/load */
	 case 'F':
	    if(page_list.pos < page_list.count-1) {    /* can go forward (not at last entry) */
	       int	pos;

	       for(pos=page_list.pos+1; pos<page_list.count-1; ++pos)    /* traverse history forward (stop at last entry) */
		  if(page_list.page[pos+1]->url->absolute)    /* found matching entry (a page before one with absolute URL) */
		     break;
	       page_list.pos=pos;    /* take found entry (or last entry, if nothing found) */

	       quit=RET_HISTORY;
	    } else    /* already at last entry */
	       flash();
	    break;

	 /* return to previous mark */
	 case 'r':
	    if(page_list.pos > 0) {    /* any URL(s) to go back */
	       int	pos;

	       for(pos=page_list.pos-1; pos>0; --pos)    /* traverse history backwards (stop at first entry) */
		  if(page_list.page[pos]->mark)    /* found a marked page */
		     break;
	       page_list.pos=pos;    /* take found entry (or first entry, if nothing found) */

	       quit=RET_HISTORY;
	    } else    /* already at first entry */
	       flash();
	    break;

	 /* forward to next mark */
	 case 'R':
	    if(page_list.pos < page_list.count-1) {    /* can go forward (not at last entry) */
	       int	pos;

	       for(pos=page_list.pos+1; pos<page_list.count-1; ++pos)    /* traverse history forward (stop at last entry) */
		  if(page_list.page[pos]->mark)    /* found a marked page */
		     break;
	       page_list.pos=pos;    /* take found entry (or first entry, if nothing found) */

	       quit=RET_HISTORY;
	    } else    /* already at last entry */
	       flash();
	    break;

	 /* set page mark */
	 case 's':
	    if(page->url->proto.type!=PT_INTERNAL)    /* normal page */
	       page->mark=1;
	    else    /* internal -> can't mark */
	       flash();
	    break;

	 /* unset page mark */
	 case 'S':
	    if(page->url->proto.type!=PT_INTERNAL)    /* normal page */
	       page->mark=0;
	    else    /* internal -> can't (un)mark */
	       flash();
	    break;

	 /* enter one command in ex mode */
	 case ':':
	    quit=RET_COMMAND;
	    break;

	 /* quit */
	 case 'q':
	    quit=RET_QUIT;
	    break;

	 /* SIGWINCH received */
	 case KEY_RESIZE:
	    getch();    /* first getch() after SIGWINCH always fails... */
	    break;

#ifdef DEBUG
	 case ERR:
	    endwin();
	    fprintf(stderr, "internal error: getch() failed (can't read keyboard input)\n");
	    exit(100);
	    break;
#endif
	 
	 /* wrong key */
	 default:
	    flash();
      }
      if(page->active_anchor>=0) {
	 page->active_anchor=-1;
	 activate_anchor(page);
      }
   } while(quit==RET_NO);

   end_fullscreen();
   return quit;    /* main program needs to know whether to quit or to enter command prompt */
}
